using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Cmt6.Exceptions;

namespace Cmt6.Models
{
    public static class EntityUtil
    {
        private static bool IsIndexer(this PropertyInfo pi)
        {
            return pi.GetIndexParameters().Length > 0;
        }

        public static IEnumerable<PropertyInfo> GetProperties(Type entityType)
        {
            var rv = entityType.GetProperties().Where(p => !p.IsIndexer());

            if (entityType.IsInterface)
            {
                rv = entityType.GetInterfaces().Aggregate(rv, (current, @interface) => current.Union(GetReadableProperties(@interface)));
            }
            else
            {
                if (entityType.BaseType != null)
                {
                    rv = rv.Union(GetReadableProperties(entityType.BaseType));
                }
            }
            return rv;
        }

        public static IEnumerable<PropertyInfo> GetReadableProperties(Type entityType)
        {
            return GetProperties(entityType).Where(p => p.CanRead);
        }

        public static IEnumerable<string> GetReadablePropertiesName(Type entityType)
        {
            return GetReadableProperties(entityType).Select(p => p.Name);
        }

        public static PropertyInfo GetProperty(Type entityType, string property)
        {
            var rv = GetProperties(entityType)
                .FirstOrDefault(p => p.Name.Equals(property, StringComparison.OrdinalIgnoreCase));
            if (rv == null)
            {
                throw new PropertyNotFoundException(entityType.FullName, property);
            }
            return rv;
        }

        //public static IEnumerable<PropertyInfo> GetPublicProperties(this IEntity entity)
        //{
        //    return entity.GetType().GetProperties();
        //}

        //public static IEnumerable<PropertyInfo> GetPublicProperties(this IEntity entity, IEnumerable<string> propertyNames)
        //{
        //    return entity.GetPublicProperties()
        //        .Where(p => propertyNames.Contains(p.Name, StringComparer.OrdinalIgnoreCase));
        //}

        //public static PropertyInfo GetPublicProperty(this IEntity entity, string propertyName)
        //{
        //    var property = entity.GetPublicProperties()
        //        .FirstOrDefault(p => p.Name.Equals(propertyName, StringComparison.OrdinalIgnoreCase));
        //    if (property == null)
        //    {
        //        throw new PropertyNotFoundException(entity.GetType().FullName, propertyName);
        //    }
        //    return property;
        //}

        public static object GetPublicPropertyValue(this IEntity entity, string propertyName)
        {
            var property = GetProperty(entity.GetType(), propertyName);
            return entity.GetPublicPropertyValue(property);
        }

        public static object GetPublicPropertyValue(this IEntity entity, PropertyInfo property)
        {
            if (!property.CanRead)
                throw new PropertyNotAccessibleException(entity.GetType().FullName, property.Name, true);
            return property.GetGetMethod().Invoke(entity, null);
        }

        public static object SetPublicPropertyValue(this IEntity entity, string propertyName, object value)
        {
            var property = GetProperty(entity.GetType(), propertyName);
            if (!property.CanWrite)
                throw new PropertyNotAccessibleException(entity.GetType().FullName, propertyName, false);
            return property.GetSetMethod().Invoke(entity, new[] { value });
        }
    }
}