using System;
using System.Collections.Generic;
using System.Data.Linq;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;
using Cmt6.Models;
using Cmt6.Models.LinqToSql;

namespace Cmt6.Repositories.LinqToSql
{
    public class ContentEntityRepository<TEntity> : EntityRepositoryBase<TEntity>
        where TEntity : class, IContentEntity
    {
        private static readonly object s_lock = new object();
        private Type m_currentType;
        private Type m_wrappedType;
        private IPropertyRepository<TEntity> m_entityPropertyRepository;

        public Type AncestorType { get; set; }
        protected DataContext DataContext { get; private set; }
        public IPropertyRepository<TEntity> EntityPropertyRepository
        {
            get { return m_entityPropertyRepository; }
            set
            {
                m_entityPropertyRepository = value;
                m_entityPropertyRepository.Added += RefreshCurrentType;
                m_entityPropertyRepository.Updated += RefreshCurrentType;
                m_entityPropertyRepository.Deleted += RefreshCurrentType;
            }
        }
        protected Type CurrentType
        {
            get
            {
                if (m_currentType == null)
                {
                    lock (s_lock)
                    {
                        if (m_currentType == null)
                        {
                            m_currentType = AncestorType;
                        }
                    }
                    RefreshCurrentType(null);
                }
                return m_currentType;
            }
            set
            {
                lock (s_lock)
                {
                    m_currentType = value;
                }
            }
        }

        protected Type WrappedType
        {
            get { return Wrap(); }
        }

        protected ITable Table
        {
            get { return DataContext.GetTable(WrappedType); }
        }

        public ContentEntityRepository(string connectionString)
        {
            DataContext = new DataContext(connectionString);
        }

        public override IEnumerable<TEntity> FindAll()
        {
            return Table.Cast<TEntity>();
        }

        public override void Add(TEntity model)
        {
            Table.InsertOnSubmit(model);
            DataContext.SubmitChanges();
        }

        public override void Delete(TEntity model)
        {
            Table.DeleteOnSubmit(model);
            DataContext.SubmitChanges();
        }

        public override void Update(TEntity model)
        {
            DataContext.SubmitChanges();
        }

        public override void AddOrUpdate(TEntity model)
        {
            var entityToUpdate = FindById(model.Id);
            if (entityToUpdate == null)
                Add(model);
            else
                Update(model);
        }
        
        protected void RefreshCurrentType(IProperty<TEntity> property)
        {
            var properties = EntityPropertyRepository.FindAll();
            var currentTypeProperties = EntityUtil.GetReadableProperties(CurrentType);
            var missedProperties = properties.Where(p => !currentTypeProperties.Select(pi => pi.Name).Contains(p.Name));
            if (missedProperties.Count() > 0)
            {
                var asmName = typeof (TEntity).Name + DateTime.Now.Ticks;
                var fileName = string.Format("{0}.dll", asmName);
                var directory = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "bin2");
                //var directory = Environment.CurrentDirectory;
                var asmBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(new AssemblyName(asmName), AssemblyBuilderAccess.RunAndSave, directory);
                var modBuilder = asmBuilder.DefineDynamicModule(asmName, fileName);
                var typeBuilder = modBuilder.DefineType(asmName, TypeAttributes.Public, CurrentType);

                //var tableAttributeData = CurrentType.GetCustomAttributesData()
                //    .First(data => data.Constructor.DeclaringType.Equals(typeof (TableAttribute)));

                //var tableAttributeBuilder = new CustomAttributeBuilder(tableAttributeData.Constructor,
                //                                                       tableAttributeData.ConstructorArguments.Select(
                //                                                           arg => arg.Value).ToArray());
                //typeBuilder.SetCustomAttribute(tableAttributeBuilder);
                //var columnAttributeCtorInfo = typeof(ColumnAttribute).GetConstructor(Type.EmptyTypes);
                //var columnAttributeBuilder = new CustomAttributeBuilder(columnAttributeCtorInfo, new object[] { });
                foreach (var missedProperty in missedProperties)
                {
                    var fieldBuilder = typeBuilder.DefineField("m_" + missedProperty.Name, missedProperty.Type,
                                                               FieldAttributes.Private);
                    var propertyBuilder = typeBuilder.DefineProperty(missedProperty.Name, PropertyAttributes.HasDefault,
                                                                     //.net 4
                                                                     CallingConventions.HasThis,
                                                                     missedProperty.Type,
                                                                     null);
                    const MethodAttributes getSetAttr = MethodAttributes.Public | MethodAttributes.SpecialName
                                                        | MethodAttributes.HideBySig | MethodAttributes.Virtual;

                    var getMethodBuilder = typeBuilder.DefineMethod("get_" + missedProperty.Name, getSetAttr,
                                                                    //.net 4
                                                                    CallingConventions.HasThis,
                                                                    missedProperty.Type, null);
                    var getMethodIl = getMethodBuilder.GetILGenerator();
                    //propertyBuilder.SetCustomAttribute(columnAttributeBuilder);
                    getMethodIl.Emit(OpCodes.Ldarg_0);
                    getMethodIl.Emit(OpCodes.Ldfld, fieldBuilder);
                    getMethodIl.Emit(OpCodes.Ret);

                    var setMethodBuilder = typeBuilder.DefineMethod("set_" + missedProperty.Name, getSetAttr,
                                                                    //.net 4
                                                                    CallingConventions.HasThis,
                                                                    null, new[] {missedProperty.Type});
                    var setMethodIl = setMethodBuilder.GetILGenerator();
                    setMethodIl.Emit(OpCodes.Ldarg_0);
                    setMethodIl.Emit(OpCodes.Ldarg_1);
                    setMethodIl.Emit(OpCodes.Stfld, fieldBuilder);
                    setMethodIl.Emit(OpCodes.Ret);

                    propertyBuilder.SetGetMethod(getMethodBuilder);
                    propertyBuilder.SetSetMethod(setMethodBuilder);
                }
                //foreach (var currentTypeProperty in currentTypeProperties)
                //{
                //    var propertyBuilder = typeBuilder.DefineProperty(currentTypeProperty.Name, PropertyAttributes.HasDefault,
                //                                                     CallingConventions.HasThis, currentTypeProperty.PropertyType,
                //                                                     new[] { currentTypeProperty.PropertyType });
                //    var columnAttributeData = currentTypeProperty.GetCustomAttributesData()
                //        .First(data => data.Constructor.DeclaringType.Equals(typeof (ColumnAttribute)));
                //    columnAttributeBuilder = new CustomAttributeBuilder(columnAttributeData.Constructor,
                //                                                        columnAttributeData.ConstructorArguments.Select(
                //                                                            arg => arg.Value).ToArray());
                //    propertyBuilder.SetCustomAttribute(columnAttributeBuilder);
                //}

                var tmpType = typeBuilder.CreateType();

                asmBuilder.Save(fileName);

                var assembly = Assembly.LoadFrom(Path.Combine(directory, fileName));
                CurrentType = assembly.GetType(asmName);
            }

            //EntityFactory.TypeInitializers[typeof(TEntity)] = () => Activator.CreateInstance(CurrentType) as IEntity;
            //Wrap();
        }

        private Type Wrap()
        {
            var wrapper = new ContentEntityWrapper<TEntity>
            {
                EntityPropertyRepository = EntityPropertyRepository
            };
            var wrappedType = wrapper.Wrap(CurrentType);
            EntityFactory.TypeInitializers[typeof(TEntity)] = () => Activator.CreateInstance(wrappedType) as IEntity;
            return wrappedType;
        }
    }
}